from libc.stdint cimport int8_t, uint8_t, int32_t, int64_t, uint32_t, uint64_t

cdef extern from "minimap.h":
	#
	# Options
	#
	ctypedef struct mm_idxopt_t:
		short k, w, flag, bucket_bits
		int64_t mini_batch_size
		uint64_t batch_size

	ctypedef struct mm_mapopt_t:
		int64_t flag
		int seed
		int sdust_thres

		int max_qlen

		int bw, bw_long
		int max_gap, max_gap_ref
		int max_frag_len
		int max_chain_skip, max_chain_iter
		int min_cnt
		int min_chain_score
		float chain_gap_scale
		float chain_skip_scale
		int rmq_size_cap, rmq_inner_dist
		int rmq_rescue_size
		float rmq_rescue_ratio

		float mask_level
		int mask_len
		float pri_ratio
		int best_n

		float alt_drop

		int a, b, q, e, q2, e2
		int transition
		int sc_ambi
		int noncan
		int junc_bonus
		int zdrop, zdrop_inv
		int end_bonus
		int min_dp_max
		int min_ksw_len
		int anchor_ext_len, anchor_ext_shift
		float max_clip_ratio

		int rank_min_len
		float rank_frac

		int pe_ori, pe_bonus

		float mid_occ_frac
		float q_occ_frac
		int32_t min_mid_occ
		int32_t mid_occ
		int32_t max_occ
		int64_t mini_batch_size
		int64_t max_sw_mat
		int64_t cap_kalloc

		const char *split_prefix

	int mm_set_opt(char *preset, mm_idxopt_t *io, mm_mapopt_t *mo)
	int mm_verbose

	#
	# Indexing
	#
	ctypedef struct mm_idx_seq_t:
		char *name
		uint64_t offset
		uint32_t len

	ctypedef struct mm_idx_bucket_t:
		pass

	ctypedef struct mm_idx_t:
		int32_t b, w, k, flag
		uint32_t n_seq
		mm_idx_seq_t *seq
		uint32_t *S
		mm_idx_bucket_t *B
		void *km
		void *h

	ctypedef struct mm_idx_reader_t:
		pass

	mm_idx_reader_t *mm_idx_reader_open(const char *fn, const mm_idxopt_t *opt, const char *fn_out)
	mm_idx_t *mm_idx_reader_read(mm_idx_reader_t *r, int n_threads)
	void mm_idx_reader_close(mm_idx_reader_t *r)
	void mm_idx_destroy(mm_idx_t *mi)
	void mm_mapopt_update(mm_mapopt_t *opt, const mm_idx_t *mi)

	int mm_idx_index_name(mm_idx_t *mi)

	#
	# Mapping (key struct defined in cmappy.h below)
	#
	ctypedef struct mm_reg1_t:
		pass

	ctypedef struct mm_tbuf_t:
		pass

	mm_tbuf_t *mm_tbuf_init()
	void mm_tbuf_destroy(mm_tbuf_t *b)
	void *mm_tbuf_get_km(mm_tbuf_t *b)
	int mm_gen_cs(void *km, char **buf, int *max_len, const mm_idx_t *mi, const mm_reg1_t *r, const char *seq, int no_iden)
	int mm_gen_MD(void *km, char **buf, int *max_len, const mm_idx_t *mi, const mm_reg1_t *r, const char *seq)

#
# Helper header (because it is hard to expose mm_reg1_t with Cython)
#
cdef extern from "cmappy.h":
	ctypedef struct mm_hitpy_t:
		const char *ctg
		int32_t ctg_start, ctg_end
		int32_t qry_start, qry_end
		int32_t blen, mlen, NM, ctg_len
		uint8_t mapq, is_primary
		int8_t strand, trans_strand
		int32_t seg_id
		int32_t n_cigar32
		uint32_t *cigar32

	void mm_reg2hitpy(const mm_idx_t *mi, mm_reg1_t *r, mm_hitpy_t *h)
	void mm_free_reg1(mm_reg1_t *r)
	mm_reg1_t *mm_map_aux(const mm_idx_t *mi, const char *seq1, const char *seq2, int *n_regs, mm_tbuf_t *b, const mm_mapopt_t *opt)
	char *mappy_fetch_seq(const mm_idx_t *mi, const char *name, int st, int en, int *l)
	mm_idx_t *mappy_idx_seq(int w, int k, int is_hpc, int bucket_bits, const char *seq, int l)

	ctypedef struct kstring_t:
		unsigned l, m
		char *s

	ctypedef struct kstream_t:
		pass

	ctypedef struct kseq_t:
		kstring_t name, comment, seq, qual
		int last_char
		kstream_t *f

	kseq_t *mm_fastx_open(const char *fn)
	void mm_fastx_close(kseq_t *ks)
	int kseq_read(kseq_t *seq)

	char *mappy_revcomp(int l, const uint8_t *seq)
	int mm_verbose_level(int v)
	void mm_reset_timer()



import os
import platform
from distutils.command.build_ext import build_ext

# from setuptools import Distribution, Extension
# from setuptools.command.build_ext import build_ext
from distutils.core import Distribution, Extension

from Cython.Build import cythonize

# Define the base path for minimap2 sources
minimap2_base = "coralsnake/minimap2"

# Define extra compile arguments based on the platform
extra_compile_args = [
    "-DHAVE_KALLOC",
    "-O3",
    "-Wno-sign-compare",
    "-Wno-unused-variable",
    "-Wno-unused-but-set-variable",
    "-Wno-unused-result",
]

include_dirs = [minimap2_base]

# Adjust compile args for ARM or x86 architectures
if platform.machine() in ["aarch64", "arm64"]:
    include_dirs.append(os.path.join(minimap2_base, "sse2neon"))
    extra_compile_args.extend(["-ftree-vectorize", "-DKSW_SSE2_ONLY", "-D__SSE2__"])
else:
    extra_compile_args.append("-msse4.1")  # Note: ancient x86_64 CPUs don't have SSE4

libraries = ["z", "m", "pthread"]


def build():
    sources = [
        "python/mappy.pyx",
        "align.c",
        "bseq.c",
        "lchain.c",
        "seed.c",
        "format.c",
        "hit.c",
        "index.c",
        "pe.c",
        "options.c",
        "ksw2_extd2_sse.c",
        "ksw2_exts2_sse.c",
        "ksw2_extz2_sse.c",
        "ksw2_ll_sse.c",
        "kalloc.c",
        "kthread.c",
        "map.c",
        "misc.c",
        "sdust.c",
        "sketch.c",
        "esterr.c",
        "splitidx.c",
    ]
    depends = [
        "minimap.h",
        "bseq.h",
        "kalloc.h",
        "kdq.h",
        "khash.h",
        "kseq.h",
        "ksort.h",
        "ksw2.h",
        "kthread.h",
        "kvec.h",
        "mmpriv.h",
        "sdust.h",
        "python/cmappy.h",
        "python/cmappy.pxd",
    ]
    sources = [os.path.join(minimap2_base, file) for file in sources]
    depends = [os.path.join(minimap2_base, file) for file in depends]

    extensions = [
        Extension(
            "mappy",
            sources=sources,
            depends=depends,
            extra_compile_args=extra_compile_args,
            include_dirs=include_dirs,
            libraries=libraries,
        )
    ]
    ext_modules = cythonize(
        extensions,
        include_path=include_dirs,
        compiler_directives={"binding": True, "language_level": 3},
    )

    distribution = Distribution(
        {
            "name": "coralsnake",
            "ext_modules": ext_modules,
            "cmdclass": {"build_ext": build_ext},
        }
    )
    distribution.package_dir = {"": "coralsnake"}

    # Define the output directory for the compiled extensions
    output_dir = "coralsnake"
    cmd = build_ext(distribution)
    cmd.build_lib = output_dir  # Direct output to coralsnake directory
    cmd.inplace = 1  # Build extensions in place
    cmd.ensure_finalized()
    cmd.run()

    # Ensure the outputs are set with the correct permissions
    for output in cmd.get_outputs():
        mode = os.stat(output).st_mode
        mode |= (
            mode & 0o444
        ) >> 2  # Make read-only by owner readable by group and others
        os.chmod(output, mode)


if __name__ == "__main__":
    build()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright © 2024 Ye Chang yech1990@gmail.com
# Distributed under terms of the GNU license.
#
# Created: 2024-06-29 19:54


import os
import pickle

import numpy as np
from ncls import NCLS32
from rich.progress import track
from xopen import xopen


# Function to parse exon data into a structured dictionary
def parse_annot(tx_file, cache=True):
    # check if pickle file exists, load it if it does
    if cache and os.path.exists(tx_file + ".pickle"):
        with open(tx_file + ".pickle", "rb") as f:
            data = pickle.load(f)
            exon_tree_by_chrom_strand = data["tree"]
            for (chromosome, strand), (
                starts,
                ends,
                rids,
            ) in exon_tree_by_chrom_strand.items():
                exon_tree_by_chrom_strand[(chromosome, strand)] = NCLS32(
                    starts, ends, rids
                )
            return exon_tree_by_chrom_strand, data["info"]

    exons_by_chrom_strand = {}
    info = {}
    with open(tx_file, "r") as f:
        names = f.readline().strip("\n").split("\t")
        chrom_idx = names.index("chrom")
        strand_idx = names.index("strand")
        spans_idx = names.index("spans")
        gene_idx = names.index("gene_id")
        transcript_idx = names.index("transcript_id")

        rid = 0
        for line in f:
            record = line.strip("\n").split("\t")
            chromosome = record[chrom_idx]
            strand = record[strand_idx]
            exon_positions = record[spans_idx]
            gene_id = record[gene_idx]
            transcript_id = record[transcript_idx]
            info[rid] = (gene_id, transcript_id)
            if (chromosome, strand) not in exons_by_chrom_strand:
                exons_by_chrom_strand[(chromosome, strand)] = []
            for exon_range in exon_positions.split(","):
                start, end = map(int, exon_range.split("-"))
                exons_by_chrom_strand[(chromosome, strand)].append((start, end, rid))
            rid += 1

    exon_tree_by_chrom_strand = {}
    for (chromosome, strand), exons in exons_by_chrom_strand.items():
        starts, ends, rid = list(zip(*exons))
        starts = np.int32(starts)
        ends = np.int32(ends)
        rids = np.int64(rid)
        exon_tree_by_chrom_strand[(chromosome, strand)] = (starts, ends, rids)
    # save the parsed data to a pickle file
    if cache:
        with open(tx_file + ".pickle", "wb") as f:
            pickle.dump(
                {"tree": exon_tree_by_chrom_strand, "info": info},
                open(tx_file + ".pickle", "wb"),
            )

    for (chromosome, strand), (starts, ends, rids) in exon_tree_by_chrom_strand.items():
        exon_tree_by_chrom_strand[(chromosome, strand)] = NCLS32(starts, ends, rids)
    return exon_tree_by_chrom_strand, info


def annot_file(
    input_file, output_file, annot_file, cols=None, keep_na=True, skip_header=False
):
    tree_by_chrom_strand, info = parse_annot(annot_file)
    if cols is None:
        cols = [0, 1, 2]
    else:
        cols = [int(i) - 1 for i in cols.split(",")]
    with xopen(input_file, "rt") as fi, xopen(output_file, "wt") as fo:
        if skip_header:
            next(fi)
        for line in track(fi, description="Processing sites"):
            line = line.strip("\n")
            records = line.split("\t")
            chromosome, position, strand = [records[i] for i in cols]
            position = int(position) - 1
            tree = tree_by_chrom_strand.get((chromosome, strand))
            founded = False
            if tree:
                for _, _, rid in tree.find_overlap(position, position + 1):
                    if info.get(rid):
                        founded = True
                        gene_id, transcript_id = info[rid]
                        fo.write(f"{line}\t{gene_id}\t{transcript_id}\n")
            if not founded and keep_na:
                fo.write(f"{line}\t.\t.\n")


if __name__ == "__main__":
    import argparse

    ap = argparse.ArgumentParser()
    ap.add_argument("input_file", help="site file")
    ap.add_argument("output_file", help="output file")
    ap.add_argument("annot_file", help="annotation file")
    args = ap.parse_args()
    annot_file(args.input_file, args.output_file, args.annot_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright © 2024 Ye Chang yech1990@gmail.com
# Distributed under terms of the GNU license.
#
# Created: 2024-06-25 14:21

import logging
from collections import defaultdict
from functools import lru_cache

import pysam

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)


def get_logger(name: str) -> logging.Logger:
    return logging.getLogger(name)


_COMP = str.maketrans("ACGTNacgtn", "TGCANtgcan")


@lru_cache(maxsize=10000)
def _python_reverse_complement(seq: str) -> str:
    return seq.translate(_COMP)[::-1]


try:
    from functools import wraps

    from . import mappy as mp

    @lru_cache(maxsize=10000)
    @wraps(mp.revcomp)
    def _c_reverse_complement(seq: str) -> str:
        return mp.revcomp(seq)

    reverse_complement = _c_reverse_complement

except ImportError:
    reverse_complement = _python_reverse_complement


class Span:
    def __init__(self, start: int, end: int):
        # 0-based
        self.start = start
        self.end = end

    def __repr__(self) -> str:
        return f"(start={self.start}, end={self.end})"


class Transcript:
    def __init__(
        self,
        gene_id: str | None = None,
        transcript_id: str | None = None,
        chrom: str = "",
        strand: str = "",
        exons: dict[str | int, Span] | None = None,
        gene_name: str | None = None,
    ):
        self.gene_id = gene_id
        self.transcript_id = transcript_id
        self.chrom = chrom
        self.strand = strand
        self.exons = {} if exons is None else exons
        # calculated feature
        self._exons_forwards = None
        self._cum_exon_lens = None
        # extra feature
        self.gene_name: str | None = gene_name
        self.start_codon: Span | None = None
        self.stop_codon: Span | None = None
        self.priority: tuple[int, int] = (10, 0)

    def add_exon(self, exon_id: str | int, exon: Span) -> None:
        self.exons[exon_id] = exon
        self._exons_forwards = None
        self._cum_exon_lens = None

    def sort_exons(self) -> None:
        self.exons = dict(
            sorted(
                self.exons.items(), key=lambda x: x[1].start, reverse=self.strand == "-"
            ),
        )
        self._exons_forwards = None
        self._cum_exon_lens = None

    def to_tsv(self, with_codon=False, with_genename=False) -> str:
        # convert into 1-based
        line = []
        for key in ["gene_id", "transcript_id", "chrom", "strand"]:
            v = getattr(self, key)
            if v is not None:
                line.append(v)
            else:
                line.append("")
        line += [",".join([f"{v.start + 1}-{v.end}" for v in self.exons.values()])]
        if with_codon:
            # use the start postion of the span
            line += [
                str(self.start_codon.start + 1) if self.start_codon else "",
                str(self.stop_codon.start + 1) if self.stop_codon else "",
            ]
        if with_genename:
            line.append(self.gene_name if self.gene_name is not None else "")
        return "\t".join(line)

    def get_seq(self, fasta: pysam.FastaFile, sort=True):
        if sort:
            self.sort_exons()
        seq = ""
        for _, v in self.exons.items():
            e = fasta.fetch(self.chrom, v.start, v.end)
            if self.strand == "-":
                e = reverse_complement(e)
            seq += e
        return seq.upper()

    @property
    def exons_forwards(self) -> list[Span]:
        if self._exons_forwards is None:
            self._exons_forwards = list(self.exons.values())
            if self.strand == "-":
                self._exons_forwards = self._exons_forwards[::-1]
        return self._exons_forwards

    @property
    def cum_exon_lens(self) -> list[int]:
        if self._cum_exon_lens is None:
            lengths = [exon.end - exon.start for exon in self.exons_forwards]
            cum_lengths = []
            total = 0
            for length in lengths:
                total += length
                cum_lengths.append(total)
            return cum_lengths
        return self._cum_exon_lens

    @property
    def length(self) -> int:
        return (self.cum_exon_lens or [0])[-1]

    def __repr__(self) -> str:
        res = []
        for key in [
            "gene_id",
            "transcript_id",
            "gene_name",
            "chrom",
            "strand",
            # "exons_forwards",
        ]:
            res.append(f"{key}={getattr(self, key)}")
        return f"Transcript({', '.join(res)})"


def load_annotation(
    annotation_file: str, with_header: bool = True
) -> dict[str, dict[str, Transcript]]:
    annot = defaultdict(dict)
    with open(annotation_file, "r") as f:
        if with_header:
            next(f)  # Skip header
        for line in f:
            fields = line.strip().split("\t")
            if len(fields) < 6:
                continue
            gene_id, transcript_id, gene_name, chrom, strand, spans = fields[:6]
            exons = {
                idx: Span(int(start) - 1, int(end))
                for idx, span in enumerate(spans.split(","), 1)
                for start, end in [span.split("-")]
            }
            transcript = Transcript(
                gene_id=gene_id,
                transcript_id=transcript_id,
                chrom=chrom,
                strand=strand,
                exons=exons,
                gene_name=gene_name,
            )
            annot[gene_id][transcript_id] = transcript
    return annot


def load_faidx(faidx_file: str) -> dict[str, int]:
    faidx = {}
    with open(faidx_file, "r") as f:
        for line in f:
            fields = line.strip().split("\t")
            faidx[fields[0]] = int(fields[1])
    return faidx



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright © 2024 Ye Chang yech1990@gmail.com
# Distributed under terms of the GNU license.
#
# Created: 2024-06-08 20:32


import os

import mappy as mp

# on forward strand example
seq1 = "TCGGGTTGCTTGGGAATGCAGCCCAAAGCGGGT"
seq2 = "AGTTTACCACCCGCTTTGGGCTGCATTCCCAAGCA"

idx = mp.Aligner(fn_idx_in="./debug.fa", preset="sr")

# q_st  q_en  strand  ctg  ctg_len  r_st  r_en  mlen  blen  mapq  cg:Z:cigar_str

for hit in idx.map(seq1, seq2=seq2, cs=True, MD=True):
    print(str(hit))
print(".......")

# reverse
for hit in idx.map(seq1, seq2=seq2[::-1], cs=True, MD=True):
    print(str(hit))
print(".......")

# complement
for hit in idx.map(seq1, seq2=mp.revcomp(seq2)[::-1], cs=True, MD=True):
    print(str(hit))
print(".......")

# rev complement complement
for hit in idx.map(seq1, seq2=mp.revcomp(seq2), cs=True, MD=True):
    print(str(hit))
print(".......")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright © 2024 Ye Chang yech1990@gmail.com
# Distributed under terms of the GNU license.
#
# Created: 2024-06-23 18:01


import bisect
from functools import lru_cache

import pysam
from rich.progress import track

from .utils import (
    Transcript,
    get_logger,
    load_annotation,
    load_faidx,
    reverse_complement,
)

LOGGER = get_logger(__name__)

COMP = str.maketrans("ACGTNacgtn", "TGCANtgcan")


@lru_cache(maxsize=10000)
def flip_flag(flag):
    if flag & 16:
        return (flag & ~16) | 32
    elif flag & 32:
        return (flag & ~32) | 16
    else:
        return flag ^ 16


@lru_cache(maxsize=10000)
def reverse_md(md: str) -> str:
    parts = []
    num = ""
    for char in md:
        if char.isdigit():
            num += char
        else:
            if num:
                parts.append(num)
                num = ""
            if char == "^":
                deletion = "^"
                while md and md[0].isalpha():
                    deletion += md[0]
                    md = md[1:]
                parts.append(deletion[::-1])
            else:
                parts.append(char.translate(COMP))
    if num:
        parts.append(num)
    return "".join(reversed(parts))


@lru_cache(maxsize=100000)
def transcript_to_genome(
    transcript_pos: int, transcript: Transcript
) -> tuple[int, int]:
    """
    transcript_pos is 0-based
    """
    if transcript_pos > transcript.length or transcript_pos < 0:
        raise ValueError("Transcript position is out of range")
    # bisect_right dose not include the right bound
    # in math notation, it is [a, b)
    exon_index = bisect.bisect_right(transcript.cum_exon_lens, transcript_pos)
    if exon_index == 0:
        return transcript.exons_forwards[0].start + transcript_pos, 0
    offset = transcript_pos - transcript.cum_exon_lens[exon_index - 1]
    return transcript.exons_forwards[exon_index].start + offset, exon_index


def remap_to_genome(
    align: pysam.AlignedSegment, transcript: Transcript, header: pysam.AlignmentHeader
) -> pysam.AlignedSegment:
    new_align = pysam.AlignedSegment(header=header)
    new_align.query_name = align.query_name
    new_align.flag = align.flag

    if align.cigartuples is None:
        raise ValueError("CIGAR string is required")

    for tag in align.get_tags():
        new_align.set_tag(*tag)
    if transcript.strand == "-":
        new_align.flag = flip_flag(align.flag)
        new_align.query_sequence = reverse_complement(align.query_sequence)
        new_align.query_qualities = (
            align.query_qualities[::-1] if align.query_qualities else None
        )
        cigartuples = align.cigartuples[::-1]
        if new_align.has_tag("MD"):
            new_align.set_tag("MD", reverse_md(new_align.get_tag("MD")))
        transcript_pos = transcript.length - align.reference_end
    else:
        new_align.query_sequence = align.query_sequence
        new_align.query_qualities = align.query_qualities
        cigartuples = align.cigartuples
        transcript_pos = align.reference_start

    genome_pos, exon_index = transcript_to_genome(transcript_pos, transcript)
    new_align.reference_name = transcript.chrom
    new_align.reference_start = genome_pos

    new_cigar = []
    for op, length in cigartuples:
        if op in (0, 2):  # Match, deletion
            while length > 0:
                current_exon = transcript.exons_forwards[exon_index]
                exon_remaining = current_exon.end - genome_pos
                if length <= exon_remaining:
                    new_cigar.append((op, length))
                    genome_pos += length
                    length = 0
                else:
                    if exon_remaining > 0:
                        new_cigar.append((op, exon_remaining))
                        genome_pos += exon_remaining
                        length -= exon_remaining
                    if exon_index + 1 < len(transcript.exons_forwards):
                        next_exon = transcript.exons_forwards[exon_index + 1]
                        intron_length = next_exon.start - current_exon.end
                        new_cigar.append((3, intron_length))
                        genome_pos = next_exon.start
                        exon_index += 1
                    else:
                        break
        else:  # Insertion or soft-clip
            new_cigar.append((op, length))

    new_align.cigartuples = new_cigar
    new_align.mapping_quality = align.mapping_quality

    return new_align


def parse_alignment(
    align: pysam.AlignedSegment,
    annot: dict[str, dict[str, Transcript]],
    genome_header: pysam.AlignmentHeader,
) -> pysam.AlignedSegment:
    if (
        align.is_unmapped
        or align.reference_name is None
        or align.reference_name not in annot
    ):
        align.reference_id = -1
        align.reference_start = -1
        align.next_reference_id = -1
        align.next_reference_start = -1
        return align

    transcript = next(iter(annot[align.reference_name].values()))
    new_align = remap_to_genome(align, transcript, genome_header)
    return new_align


def convert_bam(
    input_bam: str,
    output_bam: str,
    annotation_file: str,
    faidx_file: str,
    threads: int = 8,
    sort: bool = False,
):
    LOGGER.info("Loading annotation...")
    annot = load_annotation(annotation_file)
    LOGGER.info("Loading fasta index...")
    faidx = load_faidx(faidx_file)

    with pysam.AlignmentFile(input_bam, "rb") as in_bam:
        new_header = in_bam.header.to_dict()
        # mark as unsorted
        new_header["HD"]["SO"] = "unsorted"
        new_header["SQ"] = [
            {"SN": chrom, "LN": length} for chrom, length in faidx.items()
        ]
        genome_header = pysam.AlignmentHeader.from_dict(new_header)

        with pysam.AlignmentFile(
            output_bam,
            "wb" if output_bam.endswith(".bam") else "w",
            header=genome_header,
        ) as out_bam:
            for align in track(in_bam, description="Processing..."):
                new_align = parse_alignment(align, annot, genome_header)
                out_bam.write(new_align)

    if sort:
        # skip sorting if output is not BAM file
        if not output_bam.endswith(".bam"):
            LOGGER.info("Output is not BAM file, skip sorting and indexing")
            return
        LOGGER.info("Sorting output BAM file...")
        import shutil
        import uuid

        internal_sorted_bam = f".{str(uuid.uuid4())[:8]}_{output_bam}"
        pysam.sort("-@", str(threads), "-o", internal_sorted_bam, output_bam)
        shutil.move(internal_sorted_bam, output_bam)
        pysam.index("-@", str(threads), output_bam)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Remap transcriptome BAM to genome BAM"
    )
    parser.add_argument("-i", "--input", required=True, help="Input BAM file")
    parser.add_argument("-o", "--output", required=True, help="Output BAM file")
    parser.add_argument("-a", "--annotation", required=True, help="Annotation file")
    parser.add_argument("-f", "--faidx", required=True, help="Fasta index file")
    # sort bam file
    parser.add_argument(
        "--sort", action="store_true", help="Sort output BAM file by coordinates"
    )
    parser.add_argument(
        "-t", "--threads", type=int, default=8, help="Number of threads"
    )
    args = parser.parse_args()

    convert_bam(
        args.input, args.output, args.annotation, args.faidx, args.threads, args.sort
    )



import rich_click as click

click.rich_click.COMMAND_GROUPS = {
    "coralsnake": [
        {
            "name": "Commands",
            "commands": ["prepare", "map", "liftover", "annot"],
        },
    ]
}
click.rich_click.STYLE_OPTION = "bold green"
# click.rich_click.STYLE_COMMAND = "bold blue"


@click.group(
    invoke_without_command=False,
    help="Variant (genomic variant analysis in python)",
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.version_option(None, "-v", "--version")
@click.pass_context
def cli(ctx):
    pass


@cli.command(
    help="Extract primary transcript from gtf/gff file.",
    no_args_is_help=True,
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.option("--gtf-file", "-g", "gtf_file", help="GTF file.", required=True)
@click.option("--fasta-file", "-f", "fasta_file", help="Fasta file.", required=True)
@click.option("--output-file", "-o", "output_file", help="Output file.", required=True)
@click.option("--seq-file", "-s", "seq_file", help="Sequence file.")
@click.option(
    "--sanitize",
    "-z",
    "sanitize",
    help="Sanitize name to remove specical charaters.",
    is_flag=True,
)
@click.option(
    "--with-codon",
    "-c",
    "with_codon",
    help="Include codon in the output.",
    is_flag=True,
)
@click.option(
    "--with-genename",
    "-n",
    "with_genename",
    help="Include gene name in the output.",
    is_flag=True,
)
def prepare(
    gtf_file, fasta_file, output_file, seq_file, sanitize, with_codon, with_genename
):
    from .gtf2tx import parse_file

    parse_file(
        gtf_file, fasta_file, output_file, seq_file, sanitize, with_codon, with_genename
    )


@cli.command(
    help="Fetch genomic motif.",
    no_args_is_help=True,
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.option("--input-bam", "-i", "input_bam", help="Input bam file.", required=True)
@click.option(
    "--output-bam", "-o", "output_bam", help="Output bam file.", required=True
)
@click.option(
    "--annotation-file", "-a", "annotation_file", help="Annotation file.", required=True
)
@click.option("--faidx-file", "-f", "faidx_file", help="Faidx file.", required=True)
@click.option("--threads", "-t", "threads", help="Threads.", default=8)
@click.option("--sort", "-s", "sort", help="Sort.", is_flag=True)
def liftover(input_bam, output_bam, annotation_file, faidx_file, threads, sort):
    from .tbam2gbam import convert_bam

    convert_bam(input_bam, output_bam, annotation_file, faidx_file, threads, sort)


@cli.command(
    help="Map reads to reference genome.",
    no_args_is_help=True,
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.option("-r", "--ref-file", help="reference file", required=True)
@click.option("-1", "--r1-file", help="r1 file", required=True)
@click.option("-2", "--r2-file", help="r2 file", required=True)
# if --fwd is passed (or leave empty), fwd_lib will be True, otherwise if --rev is passed, fwd_lib will be False
@click.option("--fwd/--rev", "fwd_lib", default=True, help="forward or reverse library")
def map(ref_file, r1_file, r2_file, fwd_lib):
    from .mapping import map_file

    map_file(ref_file, r1_file, r2_file, fwd_lib)


@cli.command(
    help="Annotate tsv file.",
    no_args_is_help=True,
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.option("--input-file", "-i", "input_file", help="Input file.", required=True)
@click.option("--output-file", "-o", "output_file", help="Output file.", required=True)
@click.option(
    "--annot-file", "-a", "annot_file", help="Annotation file.", required=True
)
@click.option(
    "--cols", "-c", "cols", help="Columns of Chrom,Pos,Strand", default="1,2,3"
)
@click.option("--keep-na", "-k", "keep_na", help="Keep NA.", is_flag=True)
@click.option("--skip-header", "-H", "skip_header", help="Skip header.", is_flag=True)
def annot(input_file, output_file, annot_file, cols, keep_na, skip_header):
    from .annot import annot_file

    annot_file(input_file, output_file, annot_file, cols, keep_na, skip_header)


if __name__ == "__main__":
    cli()



try:
	from setuptools import setup, Extension
except ImportError:
	from distutils.core import setup
	from distutils.extension import Extension

import sys, platform

sys.path.append('python')

extra_compile_args = ['-DHAVE_KALLOC']
include_dirs = ["."]

if platform.machine() in ["aarch64", "arm64"]:
	include_dirs.append("sse2neon/")
	extra_compile_args.extend(['-ftree-vectorize', '-DKSW_SSE2_ONLY', '-D__SSE2__'])
else:
	extra_compile_args.append('-msse4.1') # WARNING: ancient x86_64 CPUs don't have SSE4

def readme():
	with open('python/README.rst') as f:
		return f.read()

setup(
	name = 'mappy',
	version = '2.28',
	url = 'https://github.com/lh3/minimap2',
	description = 'Minimap2 python binding',
	long_description = readme(),
	author = 'Heng Li',
	author_email = 'lh3@me.com',
	license = 'MIT',
	keywords = 'sequence-alignment',
	scripts = ['python/minimap2.py'],
	ext_modules = [Extension('mappy',
		sources = ['python/mappy.pyx', 'align.c', 'bseq.c', 'lchain.c', 'seed.c', 'format.c', 'hit.c', 'index.c', 'pe.c', 'options.c',
				   'ksw2_extd2_sse.c', 'ksw2_exts2_sse.c', 'ksw2_extz2_sse.c', 'ksw2_ll_sse.c',
				   'kalloc.c', 'kthread.c', 'map.c', 'misc.c', 'sdust.c', 'sketch.c', 'esterr.c', 'splitidx.c'],
		depends = ['minimap.h', 'bseq.h', 'kalloc.h', 'kdq.h', 'khash.h', 'kseq.h', 'ksort.h',
				   'ksw2.h', 'kthread.h', 'kvec.h', 'mmpriv.h', 'sdust.h',
				   'python/cmappy.h', 'python/cmappy.pxd'],
		extra_compile_args = extra_compile_args,
		include_dirs = include_dirs,
		libraries = ['z', 'm', 'pthread'])],
	classifiers = [
		'Development Status :: 5 - Production/Stable',
		'License :: OSI Approved :: MIT License',
		'Operating System :: POSIX',
		'Programming Language :: C',
		'Programming Language :: Cython',
		'Programming Language :: Python :: 2.7',
		'Programming Language :: Python :: 3',
		'Intended Audience :: Science/Research',
		'Topic :: Scientific/Engineering :: Bio-Informatics'],
	setup_requires=["cython"])
